----------------------------------------------------------------------------
-- Binary (square and multiply) exponentiation (exp_sq_mult.vhd)
--
-- LSB first
-- 
-- Computes the x^e mod f in GF(2**m)
-- Implements a sequential cincuit instantiang a multiplier and a squarer (combinational)
-- 
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package exp_parameters_package is
  constant N: natural := 128; --number of bits of exponentiation
end exp_parameters_package;

-----------------------------------
-- Bynary exponentiation for GF(P^M)
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.mod_f_multiplier_parameters_package.all;
use work.exp_parameters_package.all;

entity exp_sq_mult is
port (
  A: in polynomial;
  E: in std_logic_vector (N-1 downto 0);
  clk, reset, start: in std_logic; 
  B: out Polynomial;
  done: out std_logic
);
end exp_sq_mult;

architecture rtl of exp_sq_mult is
  component LSE_first_mod_f_multiplier
  port(
    a, b: in polynomial;
    clk, reset, start: in std_logic;  
    z: out polynomial;
    done: out std_logic );
  end component; 

  signal inic, shift_r, ce_c: std_logic;
  signal count: natural range 0 to N;
  type states is range 0 to 5;
  signal current_state: states;
  signal cc, bb, new_b, new_c: polynomial;
  signal ee: std_logic_vector (N-1 downto 0);
  signal start_mult, start_sq, done_mult, done_sq: std_logic;
  constant ONE_POLY: polynomial := (0 => (0 =>'1', others => '0'), others => zero_coef);

begin

  inst_mult: LSE_first_mod_f_multiplier port map (A => cc, B => bb,
     clk => clk, reset => reset, start => start_mult, Z => new_B, done => done_mult);

  inst_square: LSE_first_mod_f_multiplier port map (A => cc, B => cc,
     clk => clk, reset => reset, start => start_sq, Z => new_c, done => done_sq);

  counter: process(reset, clk)
  begin
    if reset = '1' then count <= 0;
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        count <= 0;
      elsif shift_r = '1' then
        count <= count+1; 
    end if;
    end if;
  end process counter;

  sh_reg_e: process(reset, clk)
  begin
    if reset = '1' then ee <= (others => '0');
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        ee <= e;
      elsif shift_r = '1' then
        ee <= '0' & ee(N-1 downto 1); 
    end if;
    end if;
  end process sh_reg_e;

  register_c: process(reset, clk)
  begin
    if reset = '1' then cc <= ZERO_POLY;
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        cc <= a;
      elsif shift_r = '1' then
        cc <= new_c; 
      end if;
    end if;
  end process register_c;

  register_b: process(reset, clk)
  begin
    if reset = '1' then bb <= ZERO_POLY;
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        bb <= ONE_POLY;
      elsif shift_r = '1' and ee(0) = '1' then
        bb <= new_b; 
      end if;
    end if;
  end process register_b;

  b <= bb;

  control_unit: process(clk, reset, current_state, ee(0))
  begin
    case current_state is
      when 0 to 1 => inic <= '0'; shift_r <= '0'; done <= '1'; ce_c <= '0'; start_sq <= '0'; start_mult <= '0';
      when 2 => inic <= '1'; shift_r <= '0'; done <= '0'; ce_c <= '0'; start_sq <= '0'; start_mult <= '0';
      when 3 => inic <= '0'; shift_r <= '0'; done <= '0'; ce_c <= '1'; start_sq <= '1'; start_mult <= ee(0);
      when 4 => inic <= '0'; shift_r <= '0'; done <= '0'; ce_c <= '1'; start_sq <= '0'; start_mult <= '0';
      when 5 => inic <= '0'; shift_r <= '1'; done <= '0'; ce_c <= '1'; start_sq <= '0'; start_mult <= '0';
    end case;

    if reset = '1' then current_state <= 0;
    elsif clk'event and clk = '1' then
      case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3; --capture operands
      when 3 => current_state <= 4; --start operations
      when 4 => if (done_sq = '1' and (ee(0) = '0' or done_mult = '1')) then current_state <= 5; end if;
      when 5 => if count = N-1 then current_state <= 0; else current_state <= 3; end if;
      end case;
    end if;
  end process control_unit;

end rtl;
