--------------------------------------------------------------------------------
--  Test for Polynomial exponentiation (ch5)
--
--  Generates random vectors and comparte results 
--
--
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.ALL;
use ieee.math_real.all; -- for UNIFORM, TRUNC
use work.mod_f_multiplier_parameters_package.all;
use work.exp_parameters_package.all;
USE std.textio.ALL;

ENTITY test_exp_polynom IS
END test_exp_polynom;

ARCHITECTURE behavior OF test_exp_polynom IS 

  -- Component Declaration for the Unit Under Test (UUT1)
  component LSE_first_mod_f_multiplier is
  port(
    a, b: in polynomial;
    clk, reset, start: in std_logic;  
    z: out polynomial;
    done: out std_logic );
  end component;

  component exp_sq_mult is
  port (
    A: in polynomial;
    E: in std_logic_vector (N-1 downto 0);
    clk, reset, start: in std_logic; 
    B: out Polynomial;
    done: out std_logic);
  end component;
  
    -- Internal signals
  SIGNAL clk, reset, start, done_exp1, done_exp2: std_logic;
  SIGNAL start_mult, done_mult: std_logic:= '0';
  constant DELAY : time := 100 ns;
  constant PERIOD : time := 200 ns;
  constant DUTY_CYCLE : real := 0.5;
  constant OFFSET : time := 0 ns;
  constant NUMBER_TESTS: natural := 10;
  
  signal x, y, z_exp, z_exp_minus_1, z_mult: polynomial := (others => zero_coef);
  signal exp, exp_minus_1 : std_logic_vector(N-1 downto 0) := (others => '0');

BEGIN

  -- Instantiate the Unit Under Test (UUT)
  uut1: exp_sq_mult port map ( A => x, E => exp, 
            clk => clk, reset => reset, start => start, 
            B => z_exp, done => done_exp1);

  exp_minus_1 <= exp - 1;

  uut2: exp_sq_mult port map ( A => x, E => exp_minus_1, 
            clk => clk, reset => reset, start => start, 
            B => z_exp_minus_1, done => done_exp2);
  uut3: LSE_first_mod_f_multiplier port map ( a => x, b => z_exp_minus_1, 
            clk => clk, reset => reset, start => start_mult, z => z_mult, done => done_mult);  

  PROCESS -- clock process for clk
  BEGIN
  WAIT for OFFSET;
  CLOCK_LOOP : LOOP
     clk <= '0';
     WAIT FOR (PERIOD *(1.0 - DUTY_CYCLE));
     clk <= '1';
     WAIT FOR (PERIOD * DUTY_CYCLE);
  END LOOP CLOCK_LOOP;
  END PROCESS;

  tb_proc : PROCESS --generate values
    PROCEDURE gen_polynom(X : out polynomial; s1, s2: inout Natural) IS
      VARIABLE i_x, i_p: natural;
      VARIABLE rand: real;
    BEGIN
        i_p := conv_integer(P);
        for i in 0 to M-1 loop
          UNIFORM(s1, s2, rand);
          i_x := INTEGER(TRUNC(rand * real(i_p)));
          x(i) := CONV_STD_LOGIC_VECTOR (i_x, K);
        end loop;
    END PROCEDURE;
    PROCEDURE gen_random(X : out std_logic_vector; w: natural; s1, s2: inout Natural) IS
      VARIABLE i_x, aux: integer;
      VARIABLE rand: real;
    BEGIN
        aux := W/16;
        for i in 1 to aux loop
          UNIFORM(s1, s2, rand);
          i_x := INTEGER(TRUNC(rand * real(2**16)));
          x(i*16-1 downto (i-1)*16) := CONV_STD_LOGIC_VECTOR (i_x, 16);
        end loop;
        UNIFORM(s1, s2, rand);
        i_x := INTEGER(TRUNC(rand * real(2**(w-aux*16))));
        x(w-1 downto aux*16) := CONV_STD_LOGIC_VECTOR (i_x, (w-aux*16));
    END PROCEDURE;

    VARIABLE seed1, seed2: positive; 
    VARIABLE i_x, i_y, i_p, i_z, i_yz_modp: integer;
    VARIABLE cycles, max_cycles, min_cycles, total_cycles: integer := 0;
    VARIABLE avg_cycles: real;
    VARIABLE initial_time, final_time: time;
    VARIABLE xx: polynomial ;
    VARIABLE ee: std_logic_vector(N-1 downto 0):= (others =>'0') ;
    VARIABLE ZERO: std_logic_vector(N-1 downto 0):= (others =>'0') ;

  BEGIN
    start <= '0'; reset <= '1';
    WAIT FOR PERIOD;
    reset <= '0';
    WAIT FOR PERIOD;
    
    for I in 1 to NUMBER_TESTS loop
      gen_polynom(xx, seed1, seed2);
      x <= xx;
      gen_random(ee, N, seed1, seed2);
      while (ee = ZERO) loop gen_random(ee, N, seed1, seed2);  end loop;
      exp <= ee;
      
      start <= '1'; 
      WAIT FOR PERIOD;
      start <= '0';
      wait until done_exp2 = '1';
      WAIT FOR PERIOD;
      start_mult <= '1'; 
      WAIT FOR PERIOD;
      start_mult <= '0';
      wait until done_mult = '1';
      WAIT FOR 2*PERIOD;
      
      IF ( z_exp /= z_mult) THEN 
        ASSERT (FALSE) REPORT "ERROR!!! X^M /= X^(M-1) * X" SEVERITY ERROR;
      END IF;  
 
    end loop;
    WAIT FOR DELAY;
    ASSERT (FALSE) REPORT "Simulation successful!.  Not error detected"
    SEVERITY FAILURE;
  END PROCESS;

END;
